# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.


import os
import tempfile
from typing import Optional, Sequence, Union

import cv2
import numpy as np

from geti_sdk.data_models import MediaType, Video
from geti_sdk.data_models.containers import MediaList
from geti_sdk.data_models.project import Dataset
from geti_sdk.http_session import GetiRequestException
from geti_sdk.rest_converters import MediaRESTConverter

from .media_client import BaseMediaClient


class VideoClient(BaseMediaClient[Video]):
    """
    Class to manage video uploads and downloads for a certain project
    """

    _MEDIA_TYPE = MediaType.VIDEO

    def get_all_videos(self, dataset: Optional[Dataset] = None) -> MediaList[Video]:
        """
        Get the ID's and filenames of all videos in the project, from a specific
        dataset. If no dataset is passed, videos from the training dataset will be
        returned

        :param dataset: Dataset for which to retrieve the videos. If no dataset is
            passed, videos from the training dataset are returned.
        :return: A list containing all Video's in the project
        """
        return self._get_all(dataset=dataset)

    def upload_video(
        self,
        video: Union[np.ndarray, str, os.PathLike],
        dataset: Optional[Dataset] = None,
    ) -> Video:
        """
        Upload a video file to the server. Accepts either a path to a video file, or
        a numpy array containing pixel data for video frames.

        In case a numpy array is passed, this method expects the array to be 4
        dimensional, it's dimensions shaped as: [frames, heigth, width, channels]. The
        framerate of the created video will be set to 1 fps.

        :param video: full path to the video on disk, or numpy array holding the video
            pixel data
        :param dataset: Dataset to which to upload the video. If no dataset is
            passed, the video is uploaded to the training dataset
        :return: Video object representing the uploaded video on the server
        """
        temporary_file_created = False
        if isinstance(video, (str, os.PathLike)):
            video_path = video
        elif isinstance(video, np.ndarray):
            try:
                n_frames, frame_height, frame_width, channels = video.shape
            except GetiRequestException as error:
                raise ValueError(
                    f"Invalid video input shape, expected a 4D numpy array with "
                    f"dimensions representing [frames, height, width, channels]. Got "
                    f"shape {video.shape}"
                ) from error
            file_out = tempfile.NamedTemporaryFile(suffix=".avi", delete=False)
            out = cv2.VideoWriter(
                file_out.name,
                cv2.VideoWriter_fourcc("M", "J", "P", "G"),
                1,
                (frame_width, frame_height),
            )
            for frame in video[:, ...]:
                out.write(frame)
            out.release()
            video_path = file_out.name
            temporary_file_created = True
        else:
            raise TypeError(f"Invalid video type: {type(video)}.")

        video_dict = self._upload(video_path, dataset=dataset)
        uploaded_video = MediaRESTConverter.from_dict(
            input_dict=video_dict, media_type=Video
        )
        uploaded_video._data = video_path
        uploaded_video._needs_tempfile_deletion = temporary_file_created
        return uploaded_video

    def upload_folder(
        self,
        path_to_folder: str,
        n_videos: int = -1,
        skip_if_filename_exists: bool = False,
        dataset: Optional[Dataset] = None,
    ) -> MediaList[Video]:
        """
        Upload all videos in a folder to the project. Returns the mapping of video
        filename to the unique ID assigned by Sonoma Creek.

        :param path_to_folder: Folder with videos to upload
        :param n_videos: Number of videos to upload from folder
        :param skip_if_filename_exists: Set to True to skip uploading of a video
            if a video with the same filename already exists in the project.
            Defaults to False
        :param dataset: Dataset to which to upload the video. If no dataset is
            passed, the video is uploaded to the training dataset
        :return: MediaList containing all video's in the project
        """
        return self._upload_folder(
            path_to_folder=path_to_folder,
            n_media=n_videos,
            skip_if_filename_exists=skip_if_filename_exists,
            dataset=dataset,
        )

    def download_all(self, path_to_folder: str, append_video_uid: bool = False) -> None:
        """
        Download all videos in a project to a folder on the local disk.

        :param path_to_folder: path to the folder in which the videos should be saved
        :param append_video_uid: True to append the UID of a video to the
            filename (separated from the original filename by an underscore, i.e.
            '{filename}_{video_id}'). If there are videos in the project with
            duplicate filename, this must be set to True to ensure all videos are
            downloaded. Otherwise videos with the same name will be skipped.
        """
        self._download_all(path_to_folder, append_media_uid=append_video_uid)

    def delete_videos(self, videos: Sequence[Video]) -> bool:
        """
        Delete all Video entities in `videos` from the project.

        :param videos: List of Video entities to delete
        :return: True if all videos on the list were deleted successfully,
            False otherwise
        """
        return self._delete_media(media_list=videos)



# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

import copy
from typing import ClassVar, List, Optional, Tuple

import attr
from otx.api.entities.color import Color
from otx.api.entities.color import Color as OteColor
from otx.api.entities.label import Domain as OteLabelDomain
from otx.api.entities.label import LabelEntity
from otx.api.entities.scored_label import ScoredLabel as OteScoredLabel

from geti_sdk.data_models.enums import TaskType


@attr.define
class LabelSource:
    """
    Representation of a source for a ScoredLabel in GETi

    :var user_id: ID of the user who assigned the label, if any
    :var model_id: ID of the model which generated the label, if any
    :var model_storage_id: ID of the model storage to which the model belongs
    """

    user_id: Optional[str] = None
    model_id: Optional[str] = None
    model_storage_id: Optional[str] = None
    # keys 'id' and 'type' are deprecated in v1.2, but required for v1.1
    id: Optional[str] = None
    type: Optional[str] = None


@attr.define
class Label:
    """
    Representation of a Label in GETi.

    :var name: Name of the label
    :var id: Unique database ID of the label
    :var color: Color (in hex representation) of the label
    :var group: Name of the label group to which the label belongs
    :var is_empty: True if the label represents an empty label, False otherwise
    :var parent_id: Optional name of the parent label, if any
    """

    _identifier_fields: ClassVar[List[str]] = ["id", "hotkey"]
    _GET_only_fields: ClassVar[List[str]] = ["is_empty", "is_anomalous"]

    name: str
    color: str
    group: str
    is_empty: bool
    hotkey: str = ""
    id: Optional[str] = None
    parent_id: Optional[str] = None
    is_anomalous: Optional[bool] = None

    def to_ote(self, task_type: TaskType) -> LabelEntity:
        """
        Convert the `Label` instance to an OTE SDK LabelEntity object.

        :return: OTE SDK LabelEntity instance corresponding to the label
        """
        return LabelEntity(
            name=self.name,
            domain=task_type.to_ote_domain(),
            id=self.id,
            hotkey=self.hotkey,
            is_empty=self.is_empty,
            color=Color.from_hex_str(self.color),
        )

    def prepare_for_post(self) -> None:
        """
        Set all fields to None that are not valid for making a POST request to the
        /projects endpoint.

        :return:
        """
        for field_name in self._GET_only_fields:
            setattr(self, field_name, None)

    @property
    def color_tuple(self) -> Tuple[int, int, int]:
        """
        Return the color of the label as an RGB tuple.

        :return:
        """
        hex_color_str = copy.deepcopy(self.color).strip("#")
        return tuple(int(hex_color_str[i : i + 2], 16) for i in (0, 2, 4))


@attr.define
class ScoredLabel:
    """
    Representation of a Label with an assigned probability in GETi.

    :var name: Name of the label
    :var id: Unique database ID of the label
    :var color: Color (in hex representation) of the label
    :var probability:
    :var source:
    """

    _identifier_fields: ClassVar[List[str]] = ["id"]

    probability: float
    name: Optional[str] = None
    color: Optional[str] = None
    id: Optional[str] = None
    source: Optional[LabelSource] = None

    @property
    def color_tuple(self) -> Tuple[int, int, int]:
        """
        Return the color of the label as an RGB tuple.

        :return:
        """
        hex_color_str = copy.deepcopy(self.color).strip("#")
        return tuple(int(hex_color_str[i : i + 2], 16) for i in (0, 2, 4))

    @classmethod
    def from_label(cls, label: Label, probability: float) -> "ScoredLabel":
        """
        Create a ScoredLabel instance from an input Label and probability score.

        :param label: Label to convert to ScoredLabel
        :param probability: probability score for the label
        :return: ScoredLabel instance corresponding to `label` and `probability`
        """
        return ScoredLabel(
            name=label.name, probability=probability, color=label.color, id=label.id
        )

    @classmethod
    def from_ote(cls, ote_label: OteScoredLabel) -> "ScoredLabel":
        """
        Create a :py:class`~geti_sdk.data_models.label.ScoredLabel` from
        the OTE SDK ScoredLabel entity passed.

        :param ote_label: OTE SDK ScoredLabel entity to convert from
        :return: ScoredLabel instance created according to the ote_label
        """
        return cls(
            name=ote_label.name,
            id=ote_label.id,
            probability=ote_label.probability,
            color=ote_label.color.hex_str,
        )

    def to_ote(self) -> OteScoredLabel:
        """
        Create a ScoredLabel object from OTE SDK corresponding to this
        :py:class`~geti_sdk.data_models.label.ScoredLabel` instance.
        """
        return OteScoredLabel(
            label=LabelEntity(
                name=self.name,
                color=OteColor(*self.color_tuple),
                id=self.id,
                domain=OteLabelDomain.NULL,
            ),
            probability=self.probability,
        )



# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

import abc
import math
from typing import Any, Dict, List, Optional, TypeVar, Union

import attr
import cv2
import numpy as np
from otx.api.entities.shapes.ellipse import Ellipse as OteEllipse
from otx.api.entities.shapes.polygon import Point as OtePoint
from otx.api.entities.shapes.polygon import Polygon as OtePolygon
from otx.api.entities.shapes.rectangle import Rectangle as OteRectangle
from otx.api.entities.shapes.shape import ShapeType as OteShapeType

from geti_sdk.data_models.enums import ShapeType
from geti_sdk.data_models.utils import round_to_n_digits, str_to_shape_type

# OteShapeTypeVar is a type variable that represents the possible different shapes
# that can be converted from the OTE SDK
OteShapeTypeVar = TypeVar("OteShapeTypeVar", OtePolygon, OteEllipse, OteRectangle)

# N_DIGITS_TO_ROUND_TO determines how pixel coordinates will be rounded when they are
# passed from the Intel® Geti™ REST API. The Intel® Geti™ server itself rounds some
# coordinates to 4 digits, but not all. Here we round all coordinates for internal
# consistency
N_DIGITS_TO_ROUND_TO = 0
coordinate_converter = round_to_n_digits(N_DIGITS_TO_ROUND_TO)


@attr.define(slots=False)
class Shape:
    """
    Representation of a shape in on the Intel® Geti™ platform.

    :var type: Type of the shape
    """

    type: str = attr.field(converter=str_to_shape_type)

    @abc.abstractmethod
    def to_roi(self) -> "Rectangle":
        """
        Return the bounding box containing the shape, as an instance of the Rectangle
        class.

        :return: Rectangle representing the bounding box for the shape
        """
        raise NotImplementedError

    @abc.abstractmethod
    def to_absolute_coordinates(self, parent_roi: "Rectangle") -> "Shape":
        """
        Convert the Shape to absolute coordinates, given the rectangle
        representing it's parent region of interest.

        :param parent_roi: Region of interest containing the shape
        :return: Shape converted to the coordinate system of it's parent roi
        """
        raise NotImplementedError

    @abc.abstractmethod
    def to_normalized_coordinates(
        self, image_width: int, image_height: int
    ) -> Dict[str, Any]:
        """
        Convert the Shape to a normalized coordinate system, such that all
        coordinates are represented as floats in the interval [0, 1].

        :param image_width: Width (in pixels) of the image or region with respect to
            which the shape coordinates should be normalized.
        :param image_height: Height (in pixels) of the image or region with respect to
            which the shape coordinates should be normalized
        :return: Dictionary representing the Shape object with it's coordinates
            normalized with respect to image_width and image_height
        """
        raise NotImplementedError

    @classmethod
    def from_ote(
        cls, ote_shape: OteShapeTypeVar, image_width: int, image_height: int
    ) -> Union["Rectangle", "Ellipse", "Polygon", "RotatedRectangle"]:
        """
        Create a Shape entity from a corresponding shape in the OTE SDK.

        :param ote_shape: OTE SDK shape to convert from
        :param image_width: Width of the image to which the shape applies (in pixels)
        :param image_height: Height of the image to which the shape applies (in pixels)
        :return: Shape entity created from the ote_shape
        """
        shape_mapping = {
            OteShapeType.RECTANGLE: Rectangle,
            OteShapeType.ELLIPSE: Ellipse,
            OteShapeType.POLYGON: Polygon,
        }
        return shape_mapping[ote_shape.type].from_ote(
            ote_shape, image_width=image_width, image_height=image_height
        )

    def to_ote(
        self, image_width: int, image_height: int
    ) -> Union[OtePolygon, OteEllipse, OteRectangle]:
        """
        Convert the Shape instance to the corresponding shape in the OTE SDK.

        :param image_width: Width of the image with respect to which the shape is
            defined
        :param image_height: Height of the image with respect to which the shape is
            defined
        :return: OTE SDK Rectangle, Ellipse or Polygon shape
        """
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def area(self) -> float:
        """
        Return the area of the shape, given in number of pixels.

        :return: Area of the shape, in pixels
        """
        raise NotImplementedError


@attr.define(slots=False)
class Rectangle(Shape):
    """
    Representation of a Rectangle on the Intel® Geti™ platform, as used in the
    /annotations REST endpoints.

    NOTE: All coordinates and dimensions are given in pixels

    :var x: X coordinate of the left side of the rectangle
    :var y: Y coordinate of the top of the rectangle
    :var width: Width of the rectangle
    :var height: Height of the rectangle
    """

    x: int = attr.field(converter=coordinate_converter)
    y: int = attr.field(converter=coordinate_converter)
    width: int = attr.field(converter=coordinate_converter)
    height: int = attr.field(converter=coordinate_converter)
    type: str = attr.field(
        converter=str_to_shape_type, default=ShapeType.RECTANGLE, kw_only=True
    )

    def to_normalized_coordinates(
        self, image_width: int, image_height: int
    ) -> Dict[str, float]:
        """
        Get the normalized coordinates of the rectangle, with respect to the image
        with dimensions `image_width` x `image_height`.

        :param image_width: Width of the image to which the coordinates should be
            normalized
        :param image_height: Height of the image to which the coordinates should be
            normalized
        :return: Dictionary containing the rectangle, represented in normalized
            coordinates
        """
        return dict(
            x=self.x / image_width,
            y=self.y / image_height,
            width=self.width / image_width,
            height=self.height / image_height,
            type=str(self.type),
        )

    def is_full_box(self, image_width: int, image_height: int) -> bool:
        """
        Return True if this Rectangle represents a box containing the full image.

        :param image_width: Width of the image to check for
        :param image_height: Height of the image to check for
        :return: True if the Rectangle encompasses the full image, False otherwise
        """
        return (
            self.x == 0
            and self.y == 0
            and int(self.width) == image_width
            and int(self.height) == image_height
        )

    def to_roi(self) -> "Rectangle":
        """
        Return the bounding box containing the shape, as an instance of the Rectangle
        class.

        :return: Rectangle representing the bounding box for the shape
        """
        return self

    def to_absolute_coordinates(self, parent_roi: "Rectangle") -> "Rectangle":
        """
        Convert the Rectangle to absolute coordinates, given the rectangle
        representing it's parent region of interest.

        :param parent_roi: Region of interest containing the rectangle
        :return: Rectangle converted to the coordinate system of it's parent
        """
        x_min = parent_roi.x + self.x
        y_min = parent_roi.y + self.y
        return Rectangle(x=x_min, y=y_min, width=self.width, height=self.height)

    @classmethod
    def from_ote(
        cls, ote_shape: OteRectangle, image_width: int, image_height: int
    ) -> "Rectangle":
        """
        Create a :py:class`~geti_sdk.data_models.shapes.Rectangle` from
        the OTE SDK Rectangle entity passed.

        :param ote_shape: OTE SDK Rectangle entity to convert from
        :param image_width: Width of the image to which the rectangle applies (in pixels)
        :param image_height: Height of the image to which the rectangle applies (in pixels)
        :return: Rectangle instance created according to the ote_shape
        """
        return cls(
            x=ote_shape.x1 * image_width,
            y=ote_shape.y1 * image_height,
            width=ote_shape.width * image_width,
            height=ote_shape.height * image_height,
        )

    def to_ote(self, image_width: int, image_height: int) -> OteRectangle:
        """
        Convert the Rectangle to a Rectangle instance from OTE SDK

        :param image_width: Width of the image to which the rectangle applies (in pixels)
        :param image_height: Height of the image to which the rectangle applies (in pixels)
        :return: OTE SDK Rectangle instance
        """
        return OteRectangle(
            x1=self.x / image_width,
            y1=self.y / image_height,
            x2=(self.x + self.width) / image_width,
            y2=(self.y + self.height) / image_height,
        )

    @property
    def area(self) -> float:
        """
        Return the area of the Rectangle, in pixels.

        :return:
        """
        return self.width * self.height

    @property
    def x_max(self) -> int:
        """
        Return the value of the maximal x-coordinate that the Rectangle instance
        touches.

        :return: Maximum x-coordinate for the rectangle
        """
        return self.x + self.width

    @property
    def y_max(self) -> int:
        """
        Return the value of the maximal y-coordinate that the Rectangle instance
        touches.

        :return: Maximum y-coordinate for the rectangle
        """
        return self.y + self.height


@attr.define(slots=False)
class Ellipse(Shape):
    """
    Representation of an Ellipse on the Intel® Geti™ platform, as used in the
    /annotations REST endpoints.

    NOTE: All coordinates and dimensions are given in pixels

    :var x: Lowest x coordinate of the ellipse
    :var y: Lowest y coordinate of the ellipse
    :var width: Width of the ellipse
    :var height: Height of the ellipse
    """

    x: int = attr.field(converter=coordinate_converter)
    y: int = attr.field(converter=coordinate_converter)
    width: int = attr.field(converter=coordinate_converter)
    height: int = attr.field(converter=coordinate_converter)
    type: str = attr.field(
        converter=str_to_shape_type, default=ShapeType.ELLIPSE, kw_only=True
    )

    def to_roi(self) -> "Rectangle":
        """
        Return the bounding box containing the Ellipse, as an instance of the Rectangle
        class.

        :return: Rectangle representing the bounding box for the ellipse
        """
        return Rectangle(x=self.x, y=self.y, width=self.width, height=self.height)

    def to_absolute_coordinates(self, parent_roi: Rectangle) -> "Ellipse":
        """
        Convert the Ellipse to absolute coordinates, given the rectangle
        representing it's parent region of interest.

        :param parent_roi: Region of interest containing the ellipse
        :return: Ellipse converted to the coordinate system of it's parent
        """
        x_min = parent_roi.x + self.x
        y_min = parent_roi.y + self.y
        return Ellipse(x=x_min, y=y_min, width=self.width, height=self.height)

    def to_normalized_coordinates(
        self, image_width: int, image_height: int
    ) -> Dict[str, float]:
        """
        Get the normalized coordinates of the ellipse, with respect to the image
        with dimensions `image_width` x `image_height`.

        :param image_width: Width of the image to which the coordinates should be
            normalized
        :param image_height: Height of the image to which the coordinates should be
            normalized
        :return: Dictionary containing the ellipse, represented in normalized
            coordinates
        """
        return dict(
            x=self.x / image_width,
            y=self.y / image_height,
            width=self.width / image_width,
            height=self.height / image_height,
            type=str(self.type),
        )

    @classmethod
    def from_ote(
        cls, ote_shape: OteEllipse, image_width: int, image_height: int
    ) -> "Ellipse":
        """
        Create a :py:class`~geti_sdk.data_models.shapes.Ellipse` from
        the OTE SDK Ellipse entity passed.

        :param ote_shape: OTE SDK Ellipse entity to convert from
        :param image_width: Width of the image to which the ellipse applies (in pixels)
        :param image_height: Height of the image to which the ellipse applies (in pixels)
        :return: Ellipse instance created according to the ote_shape
        """
        return cls(
            x=int(ote_shape.x1 * image_width),
            y=int(ote_shape.y1 * image_height),
            width=int(ote_shape.width * image_width),
            height=int(ote_shape.height * image_height),
        )

    def to_ote(self, image_width: int, image_height: int) -> OteEllipse:
        """
        Convert the Ellipse to an Ellipse instance from OTE SDK

        :param image_width: Width of the image to which the ellipse applies (in pixels)
        :param image_height: Height of the image to which the ellipse applies (in
            pixels)
        :return: OTE SDK Ellipse instance
        """
        return OteEllipse(
            x1=self.x / image_width,
            y1=self.y / image_height,
            x2=(self.x + self.width) / image_width,
            y2=(self.y + self.height) / image_height,
        )

    @property
    def area(self) -> float:
        """
        Return the area of the Ellipse, in pixels.

        :return:
        """
        return self.width * self.height * math.pi

    @property
    def x_max(self) -> int:
        """
        Return the value of the maximal x-coordinate that the Ellipse instance
        touches.

        :return: Maximum x-coordinate for the ellipse
        """
        return self.x + self.width

    @property
    def y_max(self) -> int:
        """
        Return the value of the maximal y-coordinate that the Ellipse instance
        touches.

        :return: Maximum y-coordinate for the ellipse
        """
        return self.y + self.height


@attr.define()
class Point:
    """
    Representation of a point on a 2D coordinate system. Used to define Polygons on
    the Intel® Geti™ platform.

    NOTE: All coordinates are defined in pixels

    :var x: X coordinate of the point
    :var y: Y coordinate of the point
    """

    x: int = attr.field(converter=coordinate_converter)
    y: int = attr.field(converter=coordinate_converter)


@attr.define(slots=False)
class Polygon(Shape):
    """
    Representation of a polygon on the Intel® Geti™ platform, as used in the
    /annotations REST endpoints.

    :var points: List of Points that make up the polygon
    """

    points: List[Point]
    type: str = attr.field(
        converter=str_to_shape_type, default=ShapeType.POLYGON, kw_only=True
    )

    def __attrs_post_init__(self):
        """
        Initialize private attributes.
        """
        self._contour: Optional[np.ndarray] = None
        self._x_max: Optional[int] = None
        self._y_max: Optional[int] = None

    def points_as_contour(self) -> np.ndarray:
        """
        Return the list of points for this Polygon as a numpy array representing
        contour points that can be plotted by openCV's drawContours function.

        NOTE: the contour is cached, to avoid going over the list of points twice. If
        the polygon is modified the cache has to be cleared. It is recommended to
        create a new polygon instead

        :return: Numpy array containing the contour
        """
        if self._contour is None:
            self._contour = np.array(
                [(int(point.x), int(point.y)) for point in self.points]
            )
        return self._contour

    def to_roi(self) -> "Rectangle":
        """
        Return the bounding box containing the Polygon, as an instance of the Rectangle
        class.

        :return: Rectangle representing the bounding box for the polygon
        """
        points_array = np.array([(point.x, point.y) for point in self.points])
        min_xy = points_array.min(axis=0)
        max_xy = points_array.max(axis=0)

        return Rectangle(
            x=min_xy[0],
            y=min_xy[1],
            width=max_xy[0] - min_xy[0],
            height=max_xy[1] - min_xy[1],
        )

    def to_absolute_coordinates(self, parent_roi: Rectangle) -> "Polygon":
        """
        Convert the Polygon to absolute coordinates, given the rectangle
        representing it's parent region of interest.

        :param parent_roi: Region of interest containing the polygon
        :return: Polygon converted to the coordinate system of it's parent ROI
        """
        absolute_points = [
            Point(x=parent_roi.x + point.x, y=parent_roi.y + point.y)
            for point in self.points
        ]
        return Polygon(points=absolute_points)

    def to_normalized_coordinates(
        self, image_width: int, image_height: int
    ) -> Dict[str, Union[List[Dict[str, float]], str]]:
        """
        Get the normalized coordinates of the polygon, with respect to the image
        with dimensions `image_width` x `image_height`.

        :param image_width: Width of the image to which the coordinates should be
            normalized
        :param image_height: Height of the image to which the coordinates should be
            normalized
        :return: Dictionary containing the polygon, represented in normalized
            coordinates
        """
        normalized_points: List[Dict[str, float]] = []
        for point in self.points:
            normalized_points.append(
                {"x": point.x / image_width, "y": point.y / image_height}
            )
        return dict(points=normalized_points, type=str(self.type))

    @classmethod
    def from_ote(
        cls, ote_shape: OtePolygon, image_width: int, image_height: int
    ) -> "Polygon":
        """
        Create a :py:class`~geti_sdk.data_models.shapes.Polygon` from
        the OTE SDK Polygon entity passed.

        :param ote_shape: OTE SDK Polygon entity to convert from
        :param image_width: Width of the image to which the polygon applies (in pixels)
        :param image_height: Height of the image to which the polygon applies (in pixels)
        :return: Polygon instance created according to the ote_shape
        """
        points = [
            Point(x=int(ote_point.x * image_width), y=int(ote_point.y * image_height))
            for ote_point in ote_shape.points
        ]
        return cls(points=points)

    def to_ote(self, image_width: int, image_height: int) -> OtePolygon:
        """
        Convert the Polygon to a Polygon instance from OTE SDK

        :param image_width: Width of the image to which the polygon applies (in pixels)
        :param image_height: Height of the image to which the polygon applies (in
            pixels)
        :return: OTE SDK Polygon instance
        """
        ote_points = [
            OtePoint(x=point.x / image_width, y=point.y / image_height)
            for point in self.points
        ]
        return OtePolygon(points=ote_points)

    @property
    def area(self) -> float:
        """
        Return the area of the Polygon, in pixels.

        :return: area of the polygon
        """
        return cv2.contourArea(self.points_as_contour())

    def _calculate_xy_max(self):
        """
        Calculate the maximum x and y coordinates that the Polyon touches, in pixels.
        """
        coord_maxes = self.points_as_contour().max(axis=0)
        self._x_max, self._y_max = coord_maxes[0], coord_maxes[1]

    @property
    def x_max(self) -> int:
        """
        Return the maximum x-coordinate of the Polygon, in pixels

        :return: largest x-coordinate that the polygon touches
        """
        if self._x_max is None:
            self._calculate_xy_max()
        return self._x_max

    @property
    def y_max(self) -> int:
        """
        Return the maximum y-coordinate of the Polygon, in pixels

        :return: largest y-coordinate that the polygon touches
        """
        if self._y_max is None:
            self._calculate_xy_max()
        return self._y_max

    def fit_rotated_rectangle(self) -> "RotatedRectangle":
        """
        Fit a RotatedRectangle object around the Polygon, such that the area spanned
        by the rectangle is minimal.

        :return: RotatedRectangle object with minimal area, which encloses the Polygon
        """
        center, (width, height), angle = cv2.minAreaRect(self.points_as_contour())
        return RotatedRectangle(
            angle=angle, x=center[0], y=center[1], width=width, height=height
        )


@attr.define(slots=False)
class RotatedRectangle(Shape):
    """
    Representation of a RotatedRectangle on the Intel® Geti™ platform, as used in the
    /annotations REST endpoints.

    NOTE: All coordinates and dimensions are specified in pixels

    :var angle: angle, in degrees, under which the rectangle is defined.
    :var x: X coordinate of the center of the rectangle
    :var y: Y coordinate of the center of the rectangle
    :var width: Width of the rectangle
    :var height: Height of the rectangle
    """

    angle: float = attr.field(converter=round_to_n_digits(n=4))
    x: int = attr.field(converter=coordinate_converter)
    y: int = attr.field(converter=coordinate_converter)
    width: int = attr.field(converter=coordinate_converter)
    height: int = attr.field(converter=coordinate_converter)
    type: str = attr.field(
        converter=str_to_shape_type, default=ShapeType.ROTATED_RECTANGLE, kw_only=True
    )

    @property
    def _angle_x_radian(self) -> float:
        """
        Return the angle to the x-axis, in radians.

        :return: angle to the x-axis in radians
        """
        return (2 * math.pi / 360) * self.angle

    @property
    def x_max(self) -> float:
        """
        Return the value of the maximal x-coordinate that the rotated rectangle touches.

        :return: Maximum x-coordinate for the rotated rectangle
        """
        return (
            self.x
            + 0.5 * self.width * math.cos(self._angle_x_radian)
            + 0.5 * self.height * math.sin(self._angle_x_radian)
        )

    @property
    def x_min(self) -> float:
        """
        Return the value of the minimal x-coordinate that the rotated rectangle touches.

        :return: Minimum x-coordinate for the rotated rectangle
        """
        return (
            self.x
            - 0.5 * self.width * math.cos(self._angle_x_radian)
            - 0.5 * self.height * math.sin(self._angle_x_radian)
        )

    @property
    def y_max(self) -> float:
        """
        Return the value of the maximal y-coordinate that the rotated rectangle touches.

        :return: Maximum y-coordinate for the rotated rectangle
        """
        return (
            self.y
            + 0.5 * self.width * math.sin(self._angle_x_radian)
            + 0.5 * self.height * math.cos(self._angle_x_radian)
        )

    @property
    def y_min(self):
        """
        Return the value of the minimal y-coordinate that the rotated rectangle touches.

        :return: Minimum y-coordinate for the rotated rectangle
        """
        return (
            self.y
            - 0.5 * self.width * math.sin(self._angle_x_radian)
            - 0.5 * self.height * math.cos(self._angle_x_radian)
        )

    @classmethod
    def from_polygon(cls, polygon: Polygon) -> "RotatedRectangle":
        """
        Create a :py:class`~geti_sdk.data_models.shapes.RotatedRectangle` from
        the Polygon entity passed.

        NOTE: The Polygon MUST consist of 4 points, otherwise a ValueError is raised

        :param polygon: Polygon entity to convert from
        :return: Rectangle instance created according to the polygon object
        """
        if len(polygon.points) != 4:
            raise ValueError(
                f"Unable to convert polygon {polygon} to RotatedRectangle. A rotated "
                f"rectangle must have exactly 4 points."
            )

        x_coords = [point.x for point in polygon.points]
        y_coords = [point.y for point in polygon.points]

        x_min, x_max = min(x_coords), max(x_coords)
        y_min, y_max = min(y_coords), max(y_coords)

        x_center = x_min + 0.5 * (x_max - abs(x_min))
        y_center = y_min + 0.5 * (y_max - abs(y_min))

        if len(x_coords) > len(set(x_coords)) or len(y_coords) > len(set(y_coords)):
            # In this case there are points sharing the same x or y value, which means
            # that we have an angle of 0 degrees
            width = x_max - x_min
            height = y_max - y_min
            alpha = 0
        else:
            x_min_coord = Point(x=x_min, y=y_coords[x_coords.index(x_min)])
            x_max_coord = Point(x=x_max, y=y_coords[x_coords.index(x_max)])
            y_min_coord = Point(x=x_coords[y_coords.index(y_min)], y=y_min)

            # Calculate the angle to the x-axis, alpha
            perpendicular_side = x_max_coord.y - y_min_coord.y
            base_side = x_max_coord.x - y_min_coord.x
            hypotenuse = math.sqrt(perpendicular_side**2 + base_side**2)

            alpha = math.asin(perpendicular_side / hypotenuse)
            alpha = alpha * 360 / (2 * math.pi)

            width = hypotenuse

            # Calculate height
            perpendicular_side_2 = x_min_coord.y - y_min_coord.y
            base_side_2 = x_min_coord.x - y_min_coord.x
            height = math.sqrt(perpendicular_side_2**2 + base_side_2**2)

        return cls(
            x=int(x_center),
            y=int(y_center),
            width=int(width),
            height=int(height),
            angle=alpha,
        )

    @classmethod
    def from_ote(
        cls, ote_shape: OtePolygon, image_width: int, image_height: int
    ) -> "RotatedRectangle":
        """
        Create a :py:class`~geti_sdk.data_models.shapes.RotatedRectangle` from
        the OTE SDK Polygon entity passed.

        NOTE: The Polygon MUST consist of 4 points, otherwise a ValueError is raised

        :param ote_shape: OTE SDK Rectangle entity to convert from
        :param image_width: Width of the image to which the shape applies (in pixels)
        :param image_height: Height of the image to which the shape applies (in pixels)
        :return: RotatedRectangle instance created according to the ote_shape
        """
        polygon = Polygon.from_ote(
            ote_shape, image_width=image_width, image_height=image_height
        )
        return cls.from_polygon(polygon)

    def to_ote(self, image_width: int, image_height: int) -> OtePolygon:
        """
        Convert the RotatedRectangle to a Polygon instance from OTE SDK

        :param image_width: Width of the image to which the rotated rectangle applies
            (in pixels)
        :param image_height: Height of the image to which the rotated rectangle
            applies (in pixels)
        :return: OTE SDK Polygon instance corresponding to the RotatedRectangle
        """
        polygon_representation = self.to_polygon()
        ote_points = [
            OtePoint(x=point.x / image_width, y=point.y / image_height)
            for point in polygon_representation.points
        ]
        return OtePolygon(points=ote_points)

    def to_roi(self) -> Rectangle:
        """
        Return the bounding box containing the RotatedRectangle, as an instance of
        the Rectangle class.

        :return: Rectangle representing the bounding box for the rotated_rectangle
        """
        return Rectangle(
            x=self.x,
            y=self.y,
            width=int(self.x_max - self.x_min),
            height=int(self.y_max - self.y_min),
        )

    def to_absolute_coordinates(self, parent_roi: Rectangle) -> "RotatedRectangle":
        """
        Convert the RotatedRectangle to absolute coordinates, given the rectangle
        representing it's parent region of interest.

        :param parent_roi: Region of interest containing the rotated rectangle
        :return: RotatedRectangle converted to the coordinate system of it's parent ROI
        """
        x = parent_roi.x + self.x
        y = parent_roi.y + self.y

        width = self.width
        height = self.height
        return RotatedRectangle(x=x, y=y, width=width, height=height, angle=self.angle)

    def to_polygon(self) -> Polygon:
        """
        Convert the RotatedRectangle instance to a Polygon consisting of 4 points.

        :return: Polygon object corresponding to the RotatedRectangle instance
        """
        y_0 = (
            self.y
            - 0.5 * self.width * math.sin(self._angle_x_radian)
            + 0.5 * self.height * math.cos(self._angle_x_radian)
        )

        x_1 = (
            self.x
            - 0.5 * self.width * math.cos(self._angle_x_radian)
            + 0.5 * self.height * math.sin(self._angle_x_radian)
        )

        y_2 = (
            self.y
            + 0.5 * self.width * math.sin(self._angle_x_radian)
            - 0.5 * self.height * math.cos(self._angle_x_radian)
        )

        x_3 = (
            self.x
            + 0.5 * self.width * math.cos(self._angle_x_radian)
            - 0.5 * self.height * math.sin(self._angle_x_radian)
        )

        point0 = Point(x=int(self.x_min), y=int(y_0))
        point1 = Point(x=int(x_1), y=int(self.y_min))
        point2 = Point(x=int(self.x_max), y=int(y_2))
        point3 = Point(x=int(x_3), y=int(self.y_max))

        return Polygon(points=[point0, point1, point2, point3])

    def to_normalized_coordinates(
        self, image_width: int, image_height: int
    ) -> Dict[str, Union[float, str]]:
        """
        Get the normalized coordinates of the rotated rectangle, with respect to the
        image with dimensions `image_width` x `image_height`.

        :param image_width: Width of the image to which the coordinates should be
            normalized
        :param image_height: Height of the image to which the coordinates should be
            normalized
        :return: Dictionary containing the rotated rectangle, represented in normalized
            coordinates
        """
        return dict(
            angle=self.angle,
            x=self.x / image_width,
            y=self.y / image_height,
            width=self.width / image_width,
            height=self.height / image_height,
            type=str(self.type),
        )

    @property
    def area(self) -> float:
        """
        Return the area of the RotatedRectangle, in pixels

        :return: area enclosed by the rotated rectangle in pixels
        """
        return self.width * self.height



# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

import json
import logging
import os
import shutil
from typing import Any, Dict, List, Optional, Union

import attr
import numpy as np

from geti_sdk.data_models import (
    Annotation,
    Label,
    Prediction,
    Project,
    ScoredLabel,
    Task,
    TaskType,
)
from geti_sdk.data_models.shapes import Polygon, Rectangle, RotatedRectangle
from geti_sdk.deployment.data_models import ROI, IntermediateInferenceResult
from geti_sdk.rest_converters import ProjectRESTConverter

from ..data_models.predictions import ResultMedium
from .deployed_model import DeployedModel
from .utils import OVMS_README_PATH, generate_ovms_model_name


@attr.define(slots=False)
class Deployment:
    """
    Representation of a deployed Intel® Geti™ project that can be used to run
    inference locally
    """

    project: Project
    models: List[DeployedModel]

    def __attrs_post_init__(self):
        """
        Initialize private attributes.
        """
        self._is_single_task: bool = len(self.project.get_trainable_tasks()) == 1
        self._are_models_loaded: bool = False
        self._inference_converters: Dict[str, Any] = {}
        self._alternate_inference_converters: Dict[str, Any] = {}
        self._empty_labels: Dict[str, Label] = {}
        self._path_to_temp_resources: Optional[str] = None
        self._requires_resource_cleanup: bool = False

    @property
    def is_single_task(self) -> bool:
        """
        Return True if the deployment represents a project with only a single task.

        :return: True if the deployed project contains only one trainable task, False
            if it is a pipeline project
        """
        return self._is_single_task

    @property
    def are_models_loaded(self) -> bool:
        """
        Return True if all inference models for the Deployment are loaded and ready
        to infer.

        :return: True if all inference models for the deployed project are loaded in
            memory and ready for inference
        """
        return self._are_models_loaded

    def save(self, path_to_folder: Union[str, os.PathLike]) -> bool:
        """
        Save the Deployment instance to a folder on local disk.

        :param path_to_folder: Folder to save the deployment to
        :return: True if the deployment was saved successfully, False otherwise
        """
        project_dict = ProjectRESTConverter.to_dict(self.project)
        deployment_folder = os.path.join(path_to_folder, "deployment")

        os.makedirs(deployment_folder, exist_ok=True, mode=0o770)

        # Save model for each task
        for task, model in zip(self.project.get_trainable_tasks(), self.models):
            model_dir = os.path.join(deployment_folder, task.title)
            os.makedirs(model_dir, exist_ok=True, mode=0o770)
            success = model.save(model_dir)
            if not success:
                logging.exception(
                    f"Saving model '{model.name}' failed. Unable to save deployment."
                )
                return False

        # Save project data
        project_filepath = os.path.join(deployment_folder, "project.json")
        with open(project_filepath, "w") as project_file:
            json.dump(project_dict, project_file, indent=4)

        # Clean up temp resources if needed
        if self._requires_resource_cleanup:
            self._remove_temporary_resources()
            self._requires_resource_cleanup = False

        return True

    @classmethod
    def from_folder(cls, path_to_folder: Union[str, os.PathLike]) -> "Deployment":
        """
        Create a Deployment instance from a specified `path_to_folder`.

        :param path_to_folder: Path to the folder containing the Deployment data
        :return: Deployment instance corresponding to the deployment data in the folder
        """
        deployment_folder = path_to_folder
        if not path_to_folder.endswith("deployment"):
            if "deployment" in os.listdir(path_to_folder):
                deployment_folder = os.path.join(path_to_folder, "deployment")
            else:
                raise ValueError(
                    f"No `deployment` folder found in the directory at "
                    f"`{path_to_folder}`. Unable to load Deployment."
                )
        project_filepath = os.path.join(deployment_folder, "project.json")
        with open(project_filepath, "r") as project_file:
            project_dict = json.load(project_file)
        project = ProjectRESTConverter.from_dict(project_dict)
        task_folder_names = [task.title for task in project.get_trainable_tasks()]
        models: List[DeployedModel] = []
        for task_folder in task_folder_names:
            models.append(
                DeployedModel.from_folder(os.path.join(deployment_folder, task_folder))
            )
        return cls(models=models, project=project)

    def load_inference_models(self, device: str = "CPU"):
        """
        Load the inference models for the deployment to the specified device.

        :param device: Device to load the inference models to
        """
        try:
            from otx.api.usecases.exportable_code.prediction_to_annotation_converter import (
                DetectionBoxToAnnotationConverter,
                IPredictionToAnnotationConverter,
                create_converter,
            )
        except ImportError as error:
            raise ValueError(
                f"Unable to load inference model for {self}. Relevant OpenVINO "
                f"packages were not found. Please make sure that all packages from the "
                f"file `requirements-deployment.txt` have been installed. "
            ) from error

        inference_converters: Dict[str, IPredictionToAnnotationConverter] = {}
        empty_labels: Dict[str, Label] = {}
        for model, task in zip(self.models, self.project.get_trainable_tasks()):
            model.load_inference_model(device=device, project=self.project)

            # This is a workaround for a bug in the label schema for anomaly tasks
            if task.type.is_anomaly:
                # For some reason the `is_anomaly` flag is not set correctly in the
                # ote_label_schema, which will break loading the prediction converter.
                # We set the flag here
                for label in model.ote_label_schema.get_labels(include_empty=True):
                    if label.name == "Anomalous":
                        label.is_anomalous = True

            inference_converter = create_converter(
                converter_type=task.type.to_ote_domain(), labels=model.ote_label_schema
            )
            inference_converters.update({task.title: inference_converter})

            # This is a workaround for a backwards incompatible change in later ote
            # versions
            if task.type.is_detection:
                alternate_inference_converter = DetectionBoxToAnnotationConverter(
                    labels=model.ote_label_schema
                )
                self._alternate_inference_converters.update(
                    {task.title: alternate_inference_converter}
                )
            empty_label = next((label for label in task.labels if label.is_empty), None)
            empty_labels.update({task.title: empty_label})

        self._inference_converters = inference_converters
        self._empty_labels = empty_labels
        self._are_models_loaded = True
        logging.info(f"Inference models loaded on device `{device}` successfully.")

    def infer(self, image: np.ndarray) -> Prediction:
        """
        Run inference on an image for the full model chain in the deployment.

        :param image: Image to run inference on, as a numpy array containing the pixel
            data. The image is expected to have dimensions [height x width x channels],
            with the channels in RGB order
        :return: inference results
        """
        self._check_models_loaded()

        # Single task inference
        if self.is_single_task:
            prediction = self._infer_task(
                image, task=self.project.get_trainable_tasks()[0], explain=False
            )
        # Multi-task inference
        else:
            prediction = self._infer_pipeline(image=image, explain=False)
        return prediction

    def explain(self, image: np.ndarray) -> Prediction:
        """
        Run inference on an image for the full model chain in the deployment. The
        resulting prediction will also contain saliency maps and the feature vector
        for the input image.

        :param image: Image to run inference on, as a numpy array containing the pixel
            data. The image is expected to have dimensions [height x width x channels],
            with the channels in RGB order
        :return: inference results
        """
        self._check_models_loaded()

        # Single task inference
        if self.is_single_task:
            prediction = self._infer_task(
                image, task=self.project.get_trainable_tasks()[0], explain=True
            )
        # Multi-task inference
        else:
            prediction = self._infer_pipeline(image=image, explain=True)
        return prediction

    def _check_models_loaded(self) -> None:
        """
        Check if models are loaded and ready for inference.

        :raises: ValueError in case models are not loaded
        """
        if not self.are_models_loaded:
            raise ValueError(
                f"Deployment '{self}' is not ready to infer, the inference models are "
                f"not loaded. Please call 'load_inference_models' first."
            )

    def _infer_task(
        self, image: np.ndarray, task: Task, explain: bool = False
    ) -> Prediction:
        """
        Run pre-processing, inference, and post-processing on the input `image`, for
        the model associated with the `task`.

        :param image: Image to run inference on
        :param task: Task to run inference for
        :param explain: True to get additional outputs for model explainability,
            including saliency maps and the feature vector for the image
        :return: Inference result
        """
        model = self._get_model_for_task(task)
        preprocessed_image, metadata = model.preprocess(image)
        inference_results = model.infer(preprocessed_image)
        postprocessing_results = model.postprocess(inference_results, metadata=metadata)

        # Optional output related to explainability
        saliency_map: Optional[np.ndarray] = None
        repr_vector: Optional[np.ndarray] = None
        if explain:
            saliency_map, repr_vector = model.postprocess_explain_outputs(
                inference_results=inference_results, metadata=metadata
            )
        converter = self._inference_converters[task.title]

        width: int = image.shape[1]
        height: int = image.shape[0]

        try:
            n_outputs = len(postprocessing_results)
        except TypeError:
            n_outputs = 1

        if n_outputs != 0:
            # The try/except is a workaround to handle different detection inference
            # results by different ote sdk versions
            try:
                annotation_scene_entity = converter.convert_to_annotation(
                    predictions=postprocessing_results, metadata=metadata
                )
            except TypeError as error:
                if task.type.is_detection:
                    converter = self._alternate_inference_converters[task.title]
                    annotation_scene_entity = converter.convert_to_annotation(
                        predictions=postprocessing_results, metadata=metadata
                    )
                    # Make sure next time we get it right in one shot
                    self._inference_converters.update({task.title: converter})
                else:
                    raise error

            prediction = Prediction.from_ote(
                annotation_scene_entity, image_width=width, image_height=height
            )
        else:
            prediction = Prediction(annotations=[])

        # Empty label is not generated by OTE correctly, append it here if there are
        # no other predictions
        if len(prediction.annotations) == 0:
            if self._empty_labels[task.title] is not None:
                prediction.append(
                    Annotation(
                        shape=Rectangle(x=0, y=0, width=width, height=height),
                        labels=[
                            ScoredLabel.from_label(
                                self._empty_labels[task.title], probability=1
                            )
                        ],
                    )
                )

        # Rotated detection models produce Polygons, convert them here to
        # RotatedRectangles
        if task.type == TaskType.ROTATED_DETECTION:
            for annotation in prediction.annotations:
                if isinstance(annotation.shape, Polygon):
                    annotation.shape = RotatedRectangle.from_polygon(annotation.shape)

        # Add optional explainability outputs
        if explain:
            prediction.feature_vector = repr_vector
            result_medium = ResultMedium(name="saliency map", type="saliency map")
            result_medium.data = saliency_map
            prediction.maps = [result_medium]

        return prediction

    def _infer_pipeline(self, image: np.ndarray, explain: bool = False) -> Prediction:
        """
        Run pre-processing, inference, and post-processing on the input `image`, for
        all models in the task chain associated with the deployment.

        Note: If `explain=True`, a saliency map, feature vector and active score for
        the first task in the pipeline will be included in the prediction output

        :param image: Image to run inference on
        :param explain: True to get additional outputs for model explainability,
            including saliency maps and the feature vector for the image
        :return: Inference result
        """
        previous_labels: Optional[List[Label]] = None
        intermediate_result: Optional[IntermediateInferenceResult] = None
        rois: Optional[List[ROI]] = None
        image_views: Optional[List[np.ndarray]] = None

        # Pipeline inference
        for task in self.project.pipeline.tasks[1:]:
            # First task in the pipeline generates the initial result and ROIs
            if task.is_trainable and previous_labels is None:
                task_prediction = self._infer_task(image, task=task, explain=explain)
                rois: Optional[List[ROI]] = None
                if not task.is_global:
                    rois = [
                        ROI.from_annotation(annotation)
                        for annotation in task_prediction.annotations
                    ]
                intermediate_result = IntermediateInferenceResult(
                    image=image, prediction=task_prediction, rois=rois
                )
                previous_labels = [label for label in task.labels if not label.is_empty]

            # Downstream trainable tasks
            elif task.is_trainable:
                if rois is None or image_views is None or intermediate_result is None:
                    raise NotImplementedError(
                        "Unable to run inference for the pipeline in the deployed "
                        "project: A flow control task is required between each "
                        "trainable task in the pipeline."
                    )
                new_rois: List[ROI] = []
                for roi, view in zip(rois, image_views):
                    view_prediction = self._infer_task(view, task)
                    if task.is_global:
                        # Global tasks add their labels to the existing shape in the ROI
                        intermediate_result.extend_annotations(
                            view_prediction.annotations, roi=roi
                        )
                    else:
                        # Local tasks create new shapes in the image coordinate system,
                        # and generate ROI's corresponding to the new shapes
                        for annotation in view_prediction.annotations:
                            intermediate_result.append_annotation(annotation, roi=roi)
                            new_rois.append(ROI.from_annotation(annotation))
                        intermediate_result.rois = [
                            new_roi.to_absolute_coordinates(parent_roi=roi)
                            for new_roi in new_rois
                        ]
                previous_labels = [label for label in task.labels if not label.is_empty]

            # Downstream flow control tasks
            else:
                if previous_labels is None:
                    raise NotImplementedError(
                        f"Unable to run inference for the pipeline in the deployed "
                        f"project: First task in the pipeline after the DATASET task "
                        f"has to be a trainable task, found task of type {task.type} "
                        f"instead."
                    )
                # CROP task
                if task.type == TaskType.CROP:
                    rois = intermediate_result.filter_rois(label=None)
                    image_views = intermediate_result.generate_views(rois)
                else:
                    raise NotImplementedError(
                        f"Unable to run inference for the pipeline in the deployed "
                        f"project: Unsupported task type {task.type} found."
                    )
        return intermediate_result.prediction

    def _get_model_for_task(self, task: Task) -> DeployedModel:
        """
        Get the DeployedModel instance corresponding to the input `task`.

        :param task: Task to get the model for
        :return: DeployedModel corresponding to the task
        """
        try:
            task_index = self.project.get_trainable_tasks().index(task)
        except ValueError as error:
            raise ValueError(
                f"Task {task.title} is not in the list of trainable tasks for project "
                f"{self.project.name}."
            ) from error
        return self.models[task_index]

    def _remove_temporary_resources(self) -> None:
        """
        If necessary, clean up any temporary resources associated with the deployment.
        """
        if self._path_to_temp_resources is not None and os.path.isdir(
            self._path_to_temp_resources
        ):
            shutil.rmtree(self._path_to_temp_resources)
        else:
            logging.debug(
                f"Unable to clean up temporary resources for deployment {self}, "
                f"because the resources were not found on the system. Possibly "
                f"they were already deleted."
            )

    def __del__(self):
        """
        If necessary, clean up any temporary resources associated with the deployment.
        This method is called when the Deployment instance is deleted.
        """
        if self._requires_resource_cleanup:
            self._remove_temporary_resources()

    def generate_ovms_config(self, output_folder: Union[str, os.PathLike]) -> None:
        """
        Generate the configuration files needed to push the models for the
        `Deployment` instance to OVMS.

        :param output_folder: Target folder to save the configuration files to
        """
        # First prepare the model config list
        if os.path.basename(output_folder) != "ovms_models":
            ovms_models_dir = os.path.join(output_folder, "ovms_models")
        else:
            ovms_models_dir = output_folder
            output_folder = os.path.dirname(ovms_models_dir)
        os.makedirs(ovms_models_dir, exist_ok=True)

        model_configs: List[Dict[str, Dict[str, Any]]] = []
        for model in self.models:
            # Create configuration entry for model
            model_name = generate_ovms_model_name(
                project=self.project, model=model, omit_version=True
            )
            config = {
                "name": model_name,
                "base_path": f"/models/{model_name}",
                "plugin_config": {"PERFORMANCE_HINT": "LATENCY"},
            }
            model_configs.append({"config": config})

            # Copy IR model files to the expected OVMS format
            if model.version is not None:
                model_version = str(model.version)
            else:
                # Fallback to version 1 if no version info is available
                model_version = "1"

            ovms_model_dir = os.path.join(ovms_models_dir, model_name, model_version)
            source_model_dir = model.model_data_path
            os.makedirs(ovms_model_dir, exist_ok=True)
            for model_file in os.listdir(source_model_dir):
                shutil.copy2(
                    src=os.path.join(source_model_dir, model_file),
                    dst=os.path.join(ovms_model_dir, model_file),
                )

        # Save model configurations
        ovms_config_list = {"model_config_list": model_configs}
        config_target_filepath = os.path.join(ovms_models_dir, "ovms_model_config.json")
        with open(config_target_filepath, "w") as file:
            json.dump(ovms_config_list, file)

        # Copy resource files
        shutil.copy2(OVMS_README_PATH, os.path.join(output_folder))

        logging.info(
            f"Configuration files for OVMS model deployment have been generated in "
            f"directory '{output_folder}'. This folder contains a `OVMS_README.md` "
            f"file with instructions on how to launch OVMS, connect to it and run "
            f"inference. Please follow the instructions outlined there to get started."
        )



# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

import re
from importlib import resources

from pathvalidate import sanitize_filepath

from geti_sdk.data_models import OptimizedModel, Project

try:
    OVMS_README_PATH = str(
        resources.files("geti_sdk.deployment.resources") / "OVMS_README.md"
    )
except AttributeError:
    with resources.path("geti_sdk.deployment.resources", "OVMS_README.md") as data_path:
        OVMS_README_PATH = str(data_path)


def generate_ovms_model_name(
    project: Project, model: OptimizedModel, omit_version: bool = True
) -> str:
    """
    Generate a valid name for a model to be uploaded to the OpenVINO Model Server.

    :param project: Project the model belongs to
    :param model: The model for which to generate the name
    :param omit_version: True not to include the version of the model in the name.
        This is useful if the name needs to be used to create a directory structure,
        since including the version will break this.
    :return: String containing the name of the model
    """
    model_name = sanitize_filepath(project.name + "_" + model.name)
    model_name = model_name.replace(" ", "_")
    model_name = model_name.replace("-", "_")
    model_name = model_name.lower()
    if model.version is not None and not omit_version:
        model_name += f":{model.version}"
    return model_name


def generate_ovms_model_address(ovms_address: str, model_name: str) -> str:
    """
    Generate a string representing a valid model address for the OpenVINO Model Server,
    from a given `ovms_address` and `model_name`.

    The format expected for model addresses by OVMS is:

        <address>:<port>/models/<model_name>[:<model_version>]

    Where `address` must not contain the protocol (http or https)

    :param ovms_address: IP address or URL pointing to the OVMS instance, including
        port specification
    :param model_name: Name of the model
    :return: String containing the model address
    """
    model_address = f"{ovms_address}/models/{model_name}"
    if model_address.startswith("https://"):
        model_address = model_address[8:]
    if model_address.startswith("http://"):
        model_address = model_address[7:]
    return model_address


def target_device_is_ovms(device: str) -> bool:
    """
    Return True if the target `device` specified is a URL or IP address, False otherwise

    :param device: Target device string to check
    :return: True if the string represents a URL or IP address, False otherwise
    """
    # Check if 'device' has been specified as a URL or IP address.
    server_pattern = re.compile(
        r"^((https?://)|(www.))(?:([a-zA-Z]+)|(\d+\.\d+\.\d+\.\d+)):\d{1,5}?$"
    )
    return server_pattern.match(device) is not None



import time
from typing import Optional, Union

from geti_sdk.data_models import Job, Task
from geti_sdk.http_session import GetiRequestException
from geti_sdk.rest_clients import TrainingClient

from .enums import SdkTestMode


def attempt_to_train_task(
    training_client: TrainingClient,
    task: Union[int, Task],
    test_mode: SdkTestMode = SdkTestMode.OFFLINE,
) -> Job:
    """
    Attempts to train the `task` (either a task or the index of the task in the list
    of trainable tasks for the project), by issuing a 'train_task' command from the
    training_client.

    If the initial train request fails with a 'project_not_train_ready' error code,
    this method will attempt to issue the command again (max 5 attempts) after
    a short sleep period.

    :param training_client: TrainingClient to use to train the task
    :param task: Task or index of task to be trained
    :param test_mode: SdkTestMode in which the training is attempted. If set to
        SdkTestMode.OFFLINE, no sleep period is used between attempts
    :return: Training job that was created by the training request
    """
    job: Optional[Job] = None
    not_ready_response = "project_not_train_ready"
    n_attempts = 10

    for i in range(n_attempts):
        try:
            job = training_client.train_task(task)
            break
        except GetiRequestException as error:
            if error.response_error_code != not_ready_response:
                raise error
        if test_mode != SdkTestMode.OFFLINE:
            time.sleep(5)

    if job is not None:
        return job
    else:
        raise ValueError(
            f"Unable to start training on server, received '{not_ready_response}' on "
            f"each request for {n_attempts} calls in a row. Please make sure that the "
            f"project is ready to start training the task."
        )



# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

from typing import Optional

from geti_sdk.data_models.containers import AlgorithmList
from geti_sdk.data_models.enums import Domain, TaskType
from geti_sdk.http_session import GetiSession


def get_supported_algorithms(
    rest_session: GetiSession,
    domain: Optional[Domain] = None,
    task_type: Optional[TaskType] = None,
) -> AlgorithmList:
    """
    Return the list of supported algorithms (including algorithm metadata) for the
    cluster.

    :param rest_session: HTTP session to the cluster
    :param domain: Optional domain for which to get the supported algorithms. If left
        as None (the default), the supported algorithms for all domains are returned.
        NOTE: domain is deprecated in SC1.1, please use `task_type` instead.
    :param task_type: Optional TaskType for which to get the supported algorithms.
    :return: AlgorithmList holding the supported algorithms
    """
    if task_type is not None and domain is not None:
        raise ValueError("Please specify either task type or domain, but not both")
    elif task_type is not None:
        query = f"?task_type={task_type}"
    elif domain is not None:
        query = f"?task_type={TaskType.from_domain(domain)}"
    else:
        query = ""

    algorithm_rest_response = rest_session.get_rest_response(
        url=f"supported_algorithms{query}", method="GET"
    )
    return AlgorithmList.from_rest(algorithm_rest_response)



# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

import datetime
import glob
import io
import os
from typing import List, Optional, Sequence, Union

import cv2
import numpy as np

from geti_sdk.data_models import Image, MediaType
from geti_sdk.data_models.containers import MediaList
from geti_sdk.data_models.project import Dataset
from geti_sdk.rest_converters import MediaRESTConverter

from .media_client import MEDIA_SUPPORTED_FORMAT_MAPPING, BaseMediaClient


class ImageClient(BaseMediaClient[Image]):
    """
    Class to manage image uploads and downloads for a certain project.
    """

    _MEDIA_TYPE = MediaType.IMAGE

    def get_all_images(self, dataset: Optional[Dataset] = None) -> MediaList[Image]:
        """
        Get the ID's and filenames of all images in the project, from a specific
        dataset. If no dataset is passed, images from the training dataset will be
        returned

        :param dataset: Dataset for which to retrieve the images. If no dataset is
            passed, images from the training dataset are returned.
        :return: MediaList containing all Image entities in the dataset
        """
        return self._get_all(dataset=dataset)

    def upload_image(
        self,
        image: Union[np.ndarray, str, os.PathLike],
        dataset: Optional[Dataset] = None,
    ) -> Image:
        """
        Upload an image file to the server.

        :param image: full path to the image on disk, or numpy array representing the
            image
        :param dataset: Dataset to which to upload the image. If no dataset is
            passed, the image is uploaded to the training dataset
        :return: Image object representing the uploaded image on the server
        """
        if isinstance(image, (str, os.PathLike)):
            image_dict = self._upload(image, dataset=dataset)
        elif isinstance(image, np.ndarray):
            image_io = io.BytesIO(cv2.imencode(".jpg", image)[1].tobytes())
            time_now = datetime.datetime.now()
            image_io.name = f"numpy_{time_now.strftime('%Y-%m-%dT%H-%M-%S.%f')}.jpg"
            image_dict = self._upload_bytes(image_io, dataset=dataset)
        else:
            raise TypeError(f"Invalid image type: {type(image)}.")
        return MediaRESTConverter.from_dict(input_dict=image_dict, media_type=Image)

    def upload_folder(
        self,
        path_to_folder: str,
        n_images: int = -1,
        skip_if_filename_exists: bool = False,
        dataset: Optional[Dataset] = None,
    ) -> MediaList[Image]:
        """
        Upload all images in a folder to the project. Returns a MediaList containing
        all images in the project after upload.

        :param path_to_folder: Folder with images to upload
        :param n_images: Number of images to upload from folder
        :param skip_if_filename_exists: Set to True to skip uploading of an image
            if an image with the same filename already exists in the project.
            Defaults to False
        :param dataset: Dataset to which to upload the images. If no dataset is
            passed, the images are uploaded to the training dataset
        :return: MediaList containing all image's in the project
        """
        return self._upload_folder(
            path_to_folder=path_to_folder,
            n_media=n_images,
            skip_if_filename_exists=skip_if_filename_exists,
            dataset=dataset,
        )

    def download_all(self, path_to_folder: str, append_image_uid: bool = False) -> None:
        """
        Download all images in a project to a folder on the local disk.

        :param path_to_folder: path to the folder in which the images should be saved
        :param append_image_uid: True to append the UID of an image to the
            filename (separated from the original filename by an underscore, i.e.
            '{filename}_{image_id}'). If there are images in the project with
            duplicate filename, this must be set to True to ensure all images are
            downloaded. Otherwise images with the same name will be skipped.
        """
        self._download_all(path_to_folder, append_media_uid=append_image_uid)

    def upload_from_list(
        self,
        path_to_folder: str,
        image_names: List[str],
        extension_included: bool = False,
        n_images: int = -1,
        skip_if_filename_exists: bool = False,
        image_names_as_full_paths: bool = False,
        dataset: Optional[Dataset] = None,
    ) -> MediaList[Image]:
        """
        From a folder containing images `path_to_folder`, this method uploads only
        those images that have their filenames included in the `image_names` list.

        :param path_to_folder: Folder containing the images
        :param image_names: List of names of the images that should be uploaded
        :param extension_included: Set to True if the extension of the image is
            included in the name, for each image in the image_names list. Defaults to
            False
        :param n_images: Number of images to upload from the list
        :param skip_if_filename_exists: Set to True to skip uploading of an image
            if an image with the same filename already exists in the project.
            Defaults to False
        :param image_names_as_full_paths: Set to True if the list of `image_names`
            contains full paths to the images, rather than just the filenames
        :param dataset: Dataset to which to upload the images. If no dataset is
            passed, the images are uploaded to the training dataset
        :return: List of images that were uploaded
        """
        media_formats = MEDIA_SUPPORTED_FORMAT_MAPPING[self._MEDIA_TYPE]

        if n_images > len(image_names) or n_images == -1:
            n_to_upload = len(image_names)
        else:
            n_to_upload = n_images

        image_filepaths: List[str] = []
        if image_names_as_full_paths:
            if extension_included:
                image_filepaths = image_names
            else:
                for image_name in image_names:
                    for media_extension in media_formats:
                        if os.path.isfile(image_name + media_extension):
                            image_filepaths.append(image_name + media_extension)
            image_filepaths = image_filepaths[0:n_to_upload]

        else:
            for image_name in image_names[0:n_to_upload]:
                if not extension_included:
                    matches: List[str] = []
                    for media_extension in media_formats:
                        matches += glob.glob(
                            os.path.join(
                                path_to_folder, "**", f"{image_name}{media_extension}"
                            ),
                            recursive=True,
                        )
                else:
                    matches = glob.glob(
                        os.path.join(path_to_folder, "**", image_name), recursive=True
                    )
                if not matches:
                    raise ValueError(
                        f"No matching file found for image with name {image_name}"
                    )
                elif len(matches) != 1:
                    raise ValueError(
                        f"Multiple files found for image with name {image_name}: "
                        f"{matches}"
                    )
                image_filepaths.append(matches[0])
        return self._upload_loop(
            filepaths=image_filepaths,
            skip_if_filename_exists=skip_if_filename_exists,
            dataset=dataset,
        )

    def delete_images(self, images: Sequence[Image]) -> bool:
        """
        Delete all Image entities in `images` from the project.

        :param images: List of Image entities to delete
        :return: True if all images on the list were deleted successfully,
            False otherwise
        """
        return self._delete_media(media_list=images)



# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.
import logging
from pprint import pformat
from typing import Any, Dict, List, Optional

import attr

from geti_sdk.data_models.enums import JobState, JobType
from geti_sdk.data_models.project import Dataset
from geti_sdk.data_models.status import StatusSummary
from geti_sdk.data_models.utils import (
    attr_value_serializer,
    str_to_datetime,
    str_to_enum_converter,
)
from geti_sdk.http_session import GetiRequestException, GetiSession


@attr.define(slots=False)
class JobStatus(StatusSummary):
    """
    Current status of a job on the Intel® Geti™ server.

    :var state: Current state of the job
    """

    state: str = attr.field(converter=str_to_enum_converter(JobState), kw_only=True)

    @classmethod
    def from_dict(cls, status_dict: Dict[str, Any]) -> "JobStatus":
        """
        Create a JobStatus object from a dictionary.

        :param status_dict: Dictionary representing a status, as returned by the
            Intel® Geti™ /status and /jobs endpoints
        :return: JobStatus object holding the status data contained in `status_dict`
        """
        return cls(**status_dict)


@attr.define
class TaskMetadata:
    """
    Metadata related to a task on the Intel® Geti™ cluster.

    :var name: Name of the task
    :var model_template_id: Identifier of the model template used by the task
    :var model_architecture: Name of the neural network architecture used for the model
    :var model_version: Version of the model currently used by the job
    :var dataset_storage_id: Unique database ID of the dataset storage used by the job
    :var task_id: ID of the task to which the TaskStatus object applies. Only used in
        Geti v1.1 and up
    """

    model_architecture: Optional[str] = None
    model_template_id: Optional[str] = None
    model_version: Optional[int] = None
    name: Optional[str] = None
    dataset_storage_id: Optional[str] = None
    task_id: Optional[str] = None  # Added in Geti v1.1


@attr.define
class TestMetadata:
    """
    Metadata related to a model test job on the GETi cluster.

    :var model_template_id: Identifier of the model template used in the test
    :var model_architecture: Name of the neural network architecture used for the model
    :var datasets: List of dictionaries, each dictionary holding the id and name of a
        dataset used in the test
    """

    model_architecture: str
    model_template_id: str
    datasets: List[Dataset]


@attr.define
class ProjectMetadata:
    """
    Metadata related to a project on the GETi cluster.

    :var name: Name of the project
    :var id: ID of the project
    """

    name: Optional[str] = None
    id: Optional[str] = None


@attr.define
class ScoreMetadata:
    """
    Metadata element containing scores for the tasks in the project

    :var task_id: ID of the task for which the score was achieved
    :var score: Performance score for the model for the task
    """

    task_id: str
    score: float


@attr.define
class JobMetadata:
    """
    Metadata for a particular job on the GETi cluster.

    :var task: TaskMetadata object holding information regarding the task from which
        the job originates
    :var base_model_id: Optional unique database ID of the base model. Only used for
        optimization jobs
    :var model_storage_id: Optional unique database ID of the model storage used by
        the job.
    :var optimization_type: Optional type of the optimization method used in the job.
        Only used for optimization jobs
    :var optimized_model_id: Optional unique database ID of the optimized model
        produced by the job. Only used for optimization jobs.
    :var scores: List of scores for the job. Added in Geti v1.1
    """

    task: Optional[TaskMetadata] = None
    project: Optional[ProjectMetadata] = None
    test: Optional[TestMetadata] = None
    base_model_id: Optional[str] = None
    model_storage_id: Optional[str] = None
    optimization_type: Optional[str] = None
    optimized_model_id: Optional[str] = None
    scores: Optional[List[ScoreMetadata]] = None


@attr.define(slots=False)
class Job:
    """
    Representation of a job running on the GETi cluster.

    :var name: Name of the job
    :var description: Description of the job
    :var id: Unique database ID of the job
    :var project_id: Unique database ID of the project from which the job originates
    :var status: JobStatus object holding the current status of the job
    :var type: Type of the job
    :var metadata: JobMetadata object holding metadata for the job
    """

    name: str
    description: str
    id: str
    status: JobStatus
    type: str = attr.field(converter=str_to_enum_converter(JobType))
    metadata: JobMetadata
    project_id: Optional[str] = None
    creation_time: Optional[str] = attr.field(converter=str_to_datetime, default=None)

    def __attrs_post_init__(self):
        """
        Initialize private attributes.
        """
        self._workspace_id: Optional[str] = None

    @property
    def workspace_id(self) -> str:
        """
        Return the unique database ID of the workspace to which the job belongs.

        :return: Unique database ID of the workspace to which the job belongs
        """
        if self._workspace_id is None:
            raise ValueError(
                f"Workspace ID for job {self} is unknown, it was never set."
            )
        return self._workspace_id

    @workspace_id.setter
    def workspace_id(self, workspace_id: str):
        """
        Set the workspace id for the job.

        :param workspace_id: Unique database ID of the workspace to which the job
            belongs
        """
        self._workspace_id = workspace_id

    @property
    def relative_url(self) -> str:
        """
        Return the url at which the Job can be addressed on the GETi cluster, relative
        to the url of the cluster itself.

        :return: Relative url for the Job instance
        """
        return f"workspaces/{self.workspace_id}/jobs/{self.id}"

    def update(self, session: GetiSession) -> "Job":
        """
        Update the job status to its current value, by making a request to the GETi
        cluster addressed by `session`.

        :param session: GetiSession to the cluster from which the Job originates
        :raises ValueError: If no workspace_id has been set for the job prior to
            calling this method
        :return: Job with its status updated
        """
        response = session.get_rest_response(url=self.relative_url, method="GET")
        updated_status = JobStatus.from_dict(response["status"])
        self.status = updated_status
        return self

    def cancel(self, session: GetiSession) -> "Job":
        """
        Cancel and delete the job, by making a request to the GETi cluster addressed
        by `session`.

        :param session: GetiSession to the cluster on which the Job is running
        :return: Job with updated status
        """
        try:
            session.get_rest_response(url=self.relative_url, method="DELETE")
            self.status.state = JobState.CANCELLED
        except GetiRequestException as error:
            if error.status_code == 404:
                logging.info(
                    f"Job '{self.name}' is not active anymore, unable to delete."
                )
                self.status.state = JobState.INACTIVE
            else:
                raise error
        return self

    @property
    def overview(self) -> str:
        """
        Return a string that shows an overview of the job.

        :return: String holding an overview of the job
        """
        return pformat(self.to_dict())

    def to_dict(self) -> Dict[str, Any]:
        """
        Return the dictionary representation of the job.

        :return:
        """
        return attr.asdict(self, recurse=True, value_serializer=attr_value_serializer)



# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.
import logging
import time
import warnings
from json import JSONDecodeError
from typing import Any, Dict, Optional, Union

import requests
import simplejson
import urllib3
from requests import Response
from requests.exceptions import RequestException
from requests.structures import CaseInsensitiveDict
from urllib3.exceptions import InsecureRequestWarning

from geti_sdk.platform_versions import GetiVersion

from .exception import GetiRequestException
from .server_config import LEGACY_API_VERSION, ServerCredentialConfig, ServerTokenConfig

CSRF_COOKIE_NAME = "_oauth2_proxy_csrf"
PROXY_COOKIE_NAME = "_oauth2_proxy"

INITIAL_HEADERS = {"Connection": "keep-alive", "Upgrade-Insecure-Requests": "1"}
SUCCESS_STATUS_CODES = [200, 201]


class GetiSession(requests.Session):
    """
    Wrapper for requests.session that sets the correct headers and cookies, and
    handles authentication and authorization.

    :param server_config: Server configuration holding the hostname (or ip address) of
        the Intel® Geti™ server, as well as the details required for authentication
        (either username and password or personal access token)
    """

    def __init__(
        self,
        server_config: Union[ServerTokenConfig, ServerCredentialConfig],
    ):
        super().__init__()
        self.headers.update(INITIAL_HEADERS)
        self.allow_redirects = False
        self.token = None
        self._cookies: Dict[str, Optional[str]] = {
            CSRF_COOKIE_NAME: None,
            PROXY_COOKIE_NAME: None,
        }

        # Configure proxies
        if server_config.proxies is None:
            self._proxies: Dict[str, str] = {}
        else:
            self._proxies = {"proxies": server_config.proxies}

        # Configure certificate verification
        if not server_config.has_valid_certificate:
            warnings.warn(
                "You have disabled TLS certificate validation, HTTPS requests made to "
                "the Intel® Geti™ server may be compromised. For optimal security, "
                "please enable certificate validation.",
                InsecureRequestWarning,
            )
            urllib3.disable_warnings(InsecureRequestWarning)
        self.verify = server_config.has_valid_certificate

        self.config = server_config
        self.logged_in = False

        # Determine authentication method
        if isinstance(server_config, ServerCredentialConfig):
            self.authenticate()
            self.use_token = False
        else:
            self.use_token = True
            access_token = self._acquire_access_token()
            self.headers.update({"Authorization": f"Bearer {access_token}"})
            self.headers.pop("Connection")

        # Get server version
        self._product_info = self._get_product_info_and_set_api_version()

    @property
    def version(self) -> GetiVersion:
        """
        Return the version of the Intel® Geti™ platform that is running on the server.

        :return: Version object holding the Intel® Geti™ version number
        """
        if "build-version" in self._product_info.keys():
            version_string = self._product_info.get("build-version")
        else:
            version_string = self._product_info.get("product-version")
        return GetiVersion(version_string)

    def _acquire_access_token(self) -> str:
        """
        Request an access token from the server, in exchange for the
        PersonalAccessToken.
        """
        try:
            response = self.get_rest_response(
                url="service_accounts/access_token",
                method="POST",
                data={"service_id": self.config.token},
                contenttype="json",
                allow_reauthentication=False,
            )
        except GetiRequestException as error:
            if error.status_code == 401:
                raise ValueError(
                    "Token authorization failed. Please make sure to provide a valid "
                    "Personal Access Token."
                )
            raise
        logging.info(f"Personal access token validated on host {self.config.host}")
        return response.get("access_token")

    def _follow_login_redirects(self, response: Response) -> str:
        """
        Recursively follow redirects in the initial login request. Updates the
        session._cookies with the cookie and the login uri.

        :param response: REST response to follow redirects for
        :return: url to the redirected location
        """
        if response.status_code in [302, 303]:
            redirect_url = response.next.url
            redirected = self.get(redirect_url, allow_redirects=False, **self._proxies)
            proxy_csrf = redirected.cookies.get(CSRF_COOKIE_NAME, None)
            if proxy_csrf:
                self._cookies[CSRF_COOKIE_NAME] = proxy_csrf
            return self._follow_login_redirects(redirected)
        else:
            return response.url

    def _get_initial_login_url(self) -> str:
        """
        Retrieve the initial login url by making a request to the login page, and
        following the redirects.

        :return: string containing the URL to the login page
        """
        response = self.get(self.config.host, allow_redirects=False, **self._proxies)
        login_page_url = self._follow_login_redirects(response)
        return login_page_url

    def authenticate(self, verbose: bool = True):
        """
        Get a new authentication cookie from the server.

        :param verbose: True to print progress output, False to suppress output
        """
        try:
            login_path = self._get_initial_login_url()
        except requests.exceptions.SSLError as error:
            raise requests.exceptions.SSLError(
                f"Connection to Intel® Geti™ server at '{self.config.host}' failed, "
                f"the server address can be resolved but the SSL certificate could not "
                f"be verified. \n Full error description: {error.args[-1]}"
            )
        except requests.exceptions.ConnectionError as error:
            if "dummy" in self.config.password or "dummy" in self.config.username:
                raise ValueError(
                    "Connection to the Intel® Geti™ server failed, please make sure to "
                    "update the user login information for the Intel® Geti™ instance."
                ) from error
            raise ValueError(
                f"Connection to the Intel® Geti™ server at host '{self.config.host}' "
                f"failed, please provide a valid cluster hostname or ip address as"
                f" well as valid login details."
            ) from error
        self.headers.clear()
        self.headers.update({"Content-Type": "application/x-www-form-urlencoded"})
        if verbose:
            logging.info(f"Authenticating on host {self.config.host}...")
        response = self.post(
            url=login_path,
            data={"login": self.config.username, "password": self.config.password},
            cookies={CSRF_COOKIE_NAME: self._cookies[CSRF_COOKIE_NAME]},
            headers={"Cookie": self._cookies[CSRF_COOKIE_NAME]},
            allow_redirects=True,
            **self._proxies,
        )
        try:
            previous_response = response.history[-1]
        except IndexError:
            raise ValueError(
                "The cluster responded to the request, but authentication failed. "
                "Please verify that you have provided correct credentials."
            )
        cookie = {PROXY_COOKIE_NAME: previous_response.cookies.get(PROXY_COOKIE_NAME)}
        self._cookies.update(cookie)
        if verbose:
            logging.info("Authentication successful. Cookie received.")
        self.logged_in = True

    def get_rest_response(
        self,
        url: str,
        method: str,
        contenttype: str = "json",
        data=None,
        allow_reauthentication: bool = True,
    ) -> Union[Response, dict, list]:
        """
        Return the REST response from a request to `url` with `method`.

        :param url: the REST url without the hostname and api pattern
        :param method: 'GET', 'POST', 'PUT', 'DELETE'
        :param contenttype: currently either 'json', 'jpeg', 'multipart', 'zip', or '',
            defaults to "json"
        :param data: the data to send in a post request, as json
        :param allow_reauthentication: True to handle authentication errors
            by attempting to re-authenticate. If set to False, such errors
            will be raised instead.
        """
        if url.startswith(self.config.api_pattern):
            url = url[len(self.config.api_pattern) :]

        if contenttype == "json":
            self.headers.update({"Content-Type": "application/json"})
        elif contenttype == "jpeg":
            self.headers.update({"Content-Type": "image/jpeg"})
        elif contenttype == "multipart":
            self.headers.pop("Content-Type", None)
        elif contenttype == "":
            self.headers.pop("Content-Type", None)
        elif contenttype == "zip":
            self.headers.update({"Content-Type": "application/zip"})

        requesturl = f"{self.config.base_url}{url}"

        if method == "POST" or method == "PUT":
            if contenttype == "json":
                kw_data_arg = {"json": data}
            elif contenttype == "multipart":
                kw_data_arg = {"files": data}
            elif contenttype == "jpeg" or contenttype == "zip":
                kw_data_arg = {"data": data}
            else:
                raise ValueError(
                    f"Making a POST request with content of type {contenttype} is "
                    f"currently not supported through the Geti SDK."
                )
        else:
            kw_data_arg = {}

        request_params = {
            "method": method,
            "url": requesturl,
            **kw_data_arg,
            "stream": True,
        }

        if not self.use_token:
            request_params.update({"cookies": self._cookies})

        try:
            response = self.request(**request_params, **self._proxies)
        except requests.exceptions.SSLError as error:
            raise requests.exceptions.SSLError(
                f"Connection to Intel® Geti™ server at '{self.config.host}' failed, "
                f"the server address can be resolved but the SSL certificate could not "
                f"be verified. \n Full error description: {error.args[-1]}"
            )

        if (
            response.status_code not in SUCCESS_STATUS_CODES
            or "text/html" in response.headers.get("Content-Type", [])
        ):
            response = self._handle_error_response(
                response=response,
                request_params=request_params,
                request_data=kw_data_arg,
                allow_reauthentication=allow_reauthentication,
            )

        if response.headers.get("Content-Type", None) == "application/json":
            result = response.json()
        else:
            result = response

        return result

    def logout(self, verbose: bool = True) -> None:
        """
        Log out of the server and end the session. All HTTPAdapters are closed and
        cookies and headers are cleared.

        :param verbose: False to suppress output messages
        """
        requires_closing = True
        requires_sign_out = True

        if not self.logged_in or self.use_token:
            requires_closing = False
            requires_sign_out = False

        if requires_sign_out:
            sign_out_url = (
                self.config.base_url[: -len(self.config.api_pattern)]
                + "/oauth2/sign_out"
            )
            try:
                response = self.request(url=sign_out_url, method="GET", **self._proxies)

                if response.status_code in SUCCESS_STATUS_CODES:
                    if verbose:
                        logging.info("Logout successful.")
                else:
                    raise GetiRequestException(
                        method="GET",
                        url=sign_out_url,
                        status_code=response.status_code,
                        request_data={},
                    )
            except (RequestException, AttributeError):
                if verbose:
                    logging.info(
                        f"The {self.__class__.__name__} is closed successfully, but "
                        f"the Geti instance was not able to logout from the server."
                    )
            except TypeError:
                # The session is already closed
                requires_closing = False

        self._cookies = {CSRF_COOKIE_NAME: None, PROXY_COOKIE_NAME: None}
        self.cookies.clear()
        self.headers = CaseInsensitiveDict(INITIAL_HEADERS)
        self.logged_in = False
        if requires_closing:
            try:
                super().close()
            except TypeError as error:
                # Sometimes a TypeError is raised if logout is called during garbage
                # collection, however we can safely ignore this since the session will
                # be deleted in that case anyway.
                logging.debug(f"{error} encountered during GetiSession closure.")

    def _get_product_info_and_set_api_version(self) -> Dict[str, str]:
        """
        Return the product info as retrieved from the Intel® Geti™ server.

        This method will also attempt to set the API version correctly, based on the
        retrieved product info.

        :return: Dictionary containing the product info.
        """
        try:
            product_info = self.get_rest_response("product_info", "GET")
        except GetiRequestException:
            self.config.api_version = LEGACY_API_VERSION
            product_info = self.get_rest_response("product_info", "GET")
        return product_info

    def __exit__(self, exc_type, exc_value, traceback):
        """
        Log out of the Intel® Geti™ server. This method is called when exiting the
        runtime context related to the session, in case the session is used as a context
        manager.
        """
        if self.logged_in:
            try:
                self.logout(verbose=False)
            except Exception as exc:
                logging.debug(f"{exc} encountered while exiting GetiSession context.")
        super().__exit__(exc_type, exc_value, traceback)

    def __del__(self):
        """
        Log out of the Intel® Geti™ server. This method is called when the session is
        deleted from memory.
        """
        if self.logged_in:
            try:
                self.logout(verbose=False)
            except Exception as exc:
                logging.debug(f"{exc} encountered during GetiSession deletion.")

    def _handle_error_response(
        self,
        response: Response,
        request_params: Dict[str, Any],
        request_data: Dict[str, Any],
        allow_reauthentication: bool = True,
    ) -> Response:
        """
        Handle error responses from the server.

        :param response: The Response object received from the server
        :param request_params: Dictionary containing the original parameters of the
            request
        :param allow_reauthentication: True to handle authentication errors
            by attempting to re-authenticate. If set to False, such errors
            will be raised instead.
        :raises: GetiRequestException in case the error cannot be handled
        :return: Response object resulting from the request
        """
        if response.status_code in [200, 401, 403] and allow_reauthentication:
            # Authentication has likely expired, re-authenticate
            logging.info("Authentication may have expired, re-authenticating...")
            if not self.use_token:
                self.authenticate(verbose=False)
                logging.info("Authentication complete.")

            else:
                access_token = self._acquire_access_token()
                logging.info("New bearer token obtained.")
                self.headers.update({"Authorization": f"Bearer {access_token}"})

            # We make one attempt to do the request again. If it fails again, a
            # GetiRequestException will be raised holding further details of the
            # reason for failure.
            response = self.request(**request_params, **self._proxies)
            if response.status_code in SUCCESS_STATUS_CODES:
                return response

        elif response.status_code == 503:
            # In case of Service Unavailable, wait some time and try again. If it
            # still doesn't work, raise exception
            time.sleep(1)
            response = self.request(**request_params, **self._proxies)

            if response.status_code in SUCCESS_STATUS_CODES:
                return response

        try:
            response_data = response.json()
        except (JSONDecodeError, simplejson.errors.JSONDecodeError):
            response_data = None

        raise GetiRequestException(
            method=request_params["method"],
            url=request_params["url"],
            status_code=response.status_code,
            request_data=request_data,
            response_data=response_data,
        )
